/*****************************************************************************/
/*  rename_dialog.c - rename dialog box and ftp routines                     */
/*  Copyright (C) 1998 Brian Masney <masneyb@newwave.net>                    */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either version 2 of the License, or        */
/*  (at your option) any later version.                                      */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                */
/*****************************************************************************/

#include "ftp.h"

static void dorenCB(GtkWidget *widget, gpointer data);
static void destroy_window(GtkWidget *widget, gpointer data);
static struct ftp_window_data *wdata;
static struct ftp_file_data *tempfle;
static GtkWidget *edit, *dialog;

void rename_dialog(GtkWidget *widget, gpointer data) {
   char tempstr[MAXSTR];
   int pos, i;
   
   wdata = (struct ftp_window_data *) data;
   if(wdata->local == -1) {
      ftp_log(LOG_MISC, "Rename: Not connected to a remote site\n");
      return;
   }
   else if(wdata->numselected != 1) {
      ftp_log(LOG_MISC, "Rename: You must only have one item selected\n");
      return;
   }

   for(i=0; i<wdata->totalitems; i++) if(wdata->items[i]) break;
   pos = i;
   tempfle = wdata->host->files;
   for(i=0; i<pos; i++) tempfle = tempfle->next;

   snprintf(tempstr, sizeof(tempstr), "What do you want to rename %s to?", tempfle->file);
   tempstr[sizeof(tempstr)-1] = '\0';
   edit = MakeEditDialog(&dialog, "Rename", tempstr, 
      "Rename", dorenCB, NULL,
      "Cancel", destroy_window, NULL);
}
/*****************************************************************************/
static void dorenCB(GtkWidget *widget, gpointer data) {
   char edttext[MAXSTR];
   int success;

   strncpy(edttext, gtk_entry_get_text(GTK_ENTRY(edit)), sizeof(edttext));
   destroy_window(widget, data);
   if(wdata->local == 1) {
      success = rename(tempfle->file, edttext) == 0;
      if(!success) {
         ftp_log(LOG_MISC, "Error: Could not rename %s to %s: %s\n", tempfle->file, edttext, sys_errlist[errno]);
      }
      else {
         ftp_log(LOG_MISC, "Successfully renamed %s to %s\n", tempfle->file, edttext);
      }
   }
   else success = ftp_rename(wdata, wdata->host, tempfle->file, edttext);
   if(success) refresh(widget, (gpointer) wdata);
}
/*****************************************************************************/
static void destroy_window(GtkWidget *widget, gpointer data) {
   gtk_grab_remove(dialog);
   gtk_widget_destroy(dialog);
}
/*****************************************************************************/
int ftp_rename(struct ftp_window_data *wdata, struct ftp_host_data *hdata, char *from, char *to) {
   if(ftp_sendcommand(1, wdata, hdata, NULL, 0, "RNFR %s\r\n", from) != '3') return(0);
   if(ftp_sendcommand(1, wdata, hdata, NULL, 0, "RNTO %s\r\n", to) != '2') return(0);
   return(1);
}
/*****************************************************************************/
